/*
 * LinkedListTemplate.h
 *
 */

#ifndef LINKEDLISTTEMPLATE_H_
#define LINKEDLISTTEMPLATE_H_

#include <iostream>

// Forward declaration of the List class
template<typename T>
class List;

// Helper class for the List class
template<typename T>
class ListNode {
private:
	ListNode(const T &nodeData, ListNode *nextPtr);

	// Data members
	T dataItem;		// List data item
	ListNode *next;	// Pointer to the next list node

	friend class List<T>;
};


/**
 * Creates a list node containing item elem and next pointer nextPtr
 */
template<typename T>
ListNode<T>::ListNode(const T &nodeDataItem, ListNode<T> *nextPtr) :
		dataItem(nodeDataItem), next(nextPtr) {
}

//--------------------------------------------------------------------

template<typename T>
class List {
public:
	List();
	~List();

	// List manipulation operations
	void insert(const T& newData);	// Insert after cursor
	bool remove();					// Remove data item
	bool replace(const T& newData);	// Replace data item
	void clear();					// Clear list

	// List status operations
	bool isEmpty() const;	// List is empty

	// List iteration operations
	bool gotoBeginning();	// Go to beginning
	bool gotoEnd();			// Go to end
	bool gotoNext();		// Go to next data item
	bool gotoPrior();		// Go to prior item
	bool getCursor(T& data) const;	// Return item

	// Output the list structure -- used in testing/debugging
	void printDebug() const;

	// Extra/Exercises
	bool moveToBeginning();					// Move to beginning
	void insertBefore(const T& newElement);	// Insert before cursor

private:
	// Data members
	ListNode<T>* head;	// Pointer to the beginning of the list
	ListNode<T>* cursor;	// Cursor pointer
};


/**
 * Creates an empty list.
 */
template<typename T>
List<T>::List() : head(0), cursor(0) {
}

template<typename T>
List<T>::~List() {
	clear();
}

/**
 * Inserts newDataItem after the cursor. If the list is empty, then
 * newDataItem is inserted as the first (and only) item in the list.
 * In either case, moves the cursor to newDataItem.
 *
 */
template<typename T>
void List<T>::insert(const T &newDataItem) {
	if (head == 0) { // Empty list
		head = new ListNode<T>(newDataItem, 0);
		cursor = head;
	} else {	// After cursor
		cursor->next = new ListNode<T>(newDataItem, cursor->next);
		cursor = cursor->next;
	}
}

/**
 * Removes the item marked by the cursor from a list. Moves the
 * cursor to the next item in the list. Assumes that the first list
 * item "follows" the last list item.
 */
template<typename T>
bool List<T>::remove() {
	ListNode<T>* p;	// Pointer to removed node
	ListNode<T>* q;	// Pointer to prior node

	// Requires that the list is not empty
	if (head == 0)
		return false;

	if (cursor == head) {
		// Remove first item
		p = head;
		head = head->next;
		cursor = head;
	} else if (cursor->next != 0) {
		// Remove middle item
		p = cursor->next;
		cursor->dataItem = p->dataItem;
		cursor->next = p->next;
	} else {
		// Remove last item
		p = cursor;
		for (q = head; q->next != cursor; q = q->next)
			;
		q->next = 0;
		cursor = head;
	}

	delete p;
	return true;
}

/**
 * Replaces the item marked by the cursor with newDataItem and
 * leaves the cursor at newDataItem.
 */
template<typename T>
bool List<T>::replace(const T &newDataItem) {
	// Requires that the list is not empty
	if (head == 0)
		return false;
	cursor->dataItem = newDataItem;
	return true;
}

/**
 * Removes all the items from a list.
 */
template<typename T>
void List<T>::clear() {
	ListNode<T>* p;	// Points to successive nodes
	ListNode<T>* nextP;	// Points to next node
	p = head;
	while (p != 0) {
		nextP = p->next;
		delete p;
		p = nextP;
	}
	head = 0;
	cursor = 0;
}


/**
 * Returns 1 if a list is empty. Otherwise, returns 0.
 */
template<typename T>
bool List<T>::isEmpty() const {
	return (head == 0);
}


/**
 * If a list is not empty, then moves the cursor to the beginning of the list
 */
template<typename T>
bool List<T>::gotoBeginning() {
	if (head==0)
		return false;
	cursor = head;
	return true;
}


/**
 * If a list is not empty, then moves the cursor to the end of the
 * list and returns 1. Otherwise, returns 0.
 */
template<typename T>
bool List<T>::gotoEnd() {
	if (head==0)
		return false;
	for (; cursor->next != 0; cursor = cursor->next)
		; // do NOTHING
	return true;
}


/**
 * If the cursor is not at the end of a list, then moves the
 * cursor to the next item in the list and returns 1. Otherwise,
 * returns 0.
 */
template<typename T>
bool List<T>::gotoNext() {
	if (cursor->next != 0) {
		cursor = cursor->next;
		return true;
	} else
		return false;
}


/**
 * If the cursor is not at the beginning of a list, then moves the
 * cursor to the preceding item in the list and returns 1.
 * Otherwise, returns 0.
 */
template<typename T>
bool List<T>::gotoPrior() {
	ListNode<T> *p;   // Pointer to prior node

	if (cursor != head) {
		for (p = head; p->next != cursor; p = p->next)
			; // do NOTHING
		cursor = p;
		return true;
	} else
		return false;
}


/**
 * Returns the item marked by the cursor.
 */
template<typename T>
bool List<T>::getCursor(T& data) const {
	// Requires that the list is not empty
	if (head == 0)
		return false;
	data = cursor->dataItem;
	return true;
}


/**
 * Outputs the items in a list. If the list is empty, outputs
 * "Empty list". This operation is intended for testing and
 * debugging purposes only.
 */
template<typename T>
void List<T>::printDebug() const {
	ListNode<T> *p;   // Iterates through the list

	if (head == 0)
		std::cout << "Empty list" << std::endl;
	else {
		for (p = head; p != 0; p = p->next)
			if (p == cursor)
				std::cout << "[" << p->dataItem << "] ";
			else
				std::cout << p->dataItem << " ";
		std::cout << std::endl;
	}
}


// Extra/Exercises
//--------------------------------------------------------------------

/**
 * Removes the item marked by the cursor from a list and
 * reinserts it at the beginning of the list. Moves the cursor to the
 * beginning of the list.
 */
template<typename T>
bool List<T>::moveToBeginning() {
	ListNode<T> *p;   //  Pointer to prior node

	// Requires that the list is not empty
	if (head == 0)
		return false;

	if (cursor != head) {
		for (p = head; p->next != cursor; p = p->next)
			; // do NOTHING
		p->next = cursor->next;
		cursor->next = head;
		head = cursor;
	}
	return true;
}


/**
 * Inserts newDataItem before the cursor. If the list is empty, then
 * newDataItem is inserted as the first (and only) item in the list.
 * In either case, moves the cursor to newDataItem.
 */
template<typename T>
void List<T>::insertBefore(const T &newDataItem) {
	if (head == 0) {	// Empty list
		head = new ListNode<T>(newDataItem, 0);
		cursor = head;
	} else { 			// Before cursor
		cursor->next = new ListNode<T>(cursor->dataItem, cursor->next);
		cursor->dataItem = newDataItem;
	}
}


#endif /* LINKEDLISTTEMPLATE_H_ */
